from typing import Optional

import mlx.core as mx
import mlx.nn as nn

from .config import VisionConfig


def check_array_shape(arr):
    shape = arr.shape

    if len(shape) != 4:
        return False

    out_channels, k_h, k_w, _ = shape
    return (out_channels >= k_h) and (out_channels >= k_w) and (k_h == k_w)


def rotate_half(x):
    """Rotates half the hidden dims of the input."""
    x1 = x[..., : x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2 :]
    return mx.concatenate([-x2, x1], axis=-1)


def apply_rotary_pos_emb_vision(tensor, freqs) -> mx.array:
    orig_dtype = tensor.dtype

    cos = mx.cos(freqs)
    sin = mx.sin(freqs)

    cos = mx.expand_dims(cos, axis=1)
    cos = mx.tile(cos, (1, 1, 2))
    cos = mx.expand_dims(cos, axis=0)

    sin = mx.expand_dims(sin, axis=1)
    sin = mx.tile(sin, (1, 1, 2))
    sin = mx.expand_dims(sin, axis=0)

    output = (tensor * cos) + (rotate_half(tensor) * sin)
    return output.astype(orig_dtype)


class VisionRotaryEmbedding(nn.Module):
    def __init__(self, dim: int, theta: float = 10000.0) -> None:
        super().__init__()
        self.dim = dim
        self.theta = theta

    def __call__(self, seqlen: int) -> mx.array:
        inv_freq = 1.0 / (
            self.theta ** (mx.arange(0, self.dim, 2, dtype=mx.float32) / self.dim)
        )
        seq = mx.arange(int(seqlen), dtype=inv_freq.dtype)
        return mx.outer(seq, inv_freq)


class RMSNorm(nn.Module):
    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.weight = mx.ones((dim,))
        self.eps = eps

    def __call__(self, x: mx.array) -> mx.array:
        input_dtype = x.dtype
        x = x.astype(mx.float32)
        x = x * mx.rsqrt(mx.mean(x**2, axis=-1, keepdims=True) + self.eps)
        return x.astype(input_dtype) * self.weight.astype(input_dtype)


class PatchMerger(nn.Module):
    def __init__(
        self,
        dim: int,
        context_dim: int,
        spatial_merge_size: int = 2,
    ) -> None:
        super().__init__()
        self.hidden_size = context_dim * (spatial_merge_size**2)
        self.ln_q = nn.LayerNorm(context_dim, eps=1e-6)
        self.mlp = [
            nn.Linear(self.hidden_size, self.hidden_size),
            nn.GELU(),
            nn.Linear(self.hidden_size, dim),
        ]

    def __call__(self, x: mx.array) -> mx.array:
        x = self.ln_q(x).reshape(-1, self.hidden_size)
        for layer in self.mlp:
            x = layer(x)
        return x


class VisionAttention(nn.Module):
    def __init__(self, config: VisionConfig) -> None:
        super().__init__()
        self.num_heads = config.num_attention_heads
        self.head_dim = config.embed_dim // config.num_attention_heads
        self.scale = self.head_dim**-0.5

        self.qkv = nn.Linear(
            config.embed_dim, config.embed_dim * 3, bias=config.use_bias
        )
        self.proj = nn.Linear(config.embed_dim, config.embed_dim, bias=config.use_bias)

    def __call__(
        self,
        hidden_states: mx.array,
        cu_seqlens: mx.array,
        rotary_pos_emb: mx.array,
    ) -> mx.array:
        seq_length = hidden_states.shape[0]

        qkv = self.qkv(hidden_states).reshape(seq_length, 3, self.num_heads, -1)
        qkv = qkv.transpose(1, 0, 2, 3)
        q, k, v = mx.split(qkv, 3)

        q = apply_rotary_pos_emb_vision(mx.expand_dims(q, 0), rotary_pos_emb)[0]
        k = apply_rotary_pos_emb_vision(mx.expand_dims(k, 0), rotary_pos_emb)[0]

        q = q.transpose(0, 2, 1, 3)
        k = k.transpose(0, 2, 1, 3)
        v = v.transpose(0, 2, 1, 3)

        splits = [
            mx.split(tensor, cu_seqlens[1:-1].tolist(), axis=2) for tensor in (q, k, v)
        ]

        attn_outputs = []
        for q_chunk, k_chunk, v_chunk in zip(*splits):
            output = mx.fast.scaled_dot_product_attention(
                q_chunk, k_chunk, v_chunk, scale=self.scale
            )
            attn_outputs.append(output)

        output = mx.concatenate(attn_outputs, axis=2)
        output = output.transpose(0, 2, 1, 3).reshape(seq_length, -1)
        return self.proj(output)


class DotsSwiGLUFFN(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()
        self.fc1 = nn.Linear(
            config.embed_dim, config.intermediate_size, bias=config.use_bias
        )
        self.fc2 = nn.Linear(
            config.intermediate_size, config.embed_dim, bias=config.use_bias
        )
        self.fc3 = nn.Linear(
            config.embed_dim, config.intermediate_size, bias=config.use_bias
        )

    def __call__(self, x: mx.array) -> mx.array:
        return self.fc2(nn.silu(self.fc1(x)) * self.fc3(x))


class DotsPatchEmbed(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()
        self.num_channels = config.num_channels
        self.patch_size = config.patch_size
        self.temporal_patch_size = config.temporal_patch_size
        self.embed_dim = config.embed_dim
        self.proj = nn.Conv2d(
            config.num_channels,
            config.embed_dim,
            kernel_size=(config.patch_size, config.patch_size),
            stride=(config.patch_size, config.patch_size),
            bias=True,
        )
        self.norm = RMSNorm(config.embed_dim, eps=config.rms_norm_eps)

    def __call__(self, x: mx.array, grid_thw=None) -> mx.array:
        x = x.reshape(
            -1,
            self.num_channels,
            self.temporal_patch_size,
            self.patch_size,
            self.patch_size,
        )[:, :, 0]
        x = x.transpose(0, 2, 3, 1)
        x = self.proj(x).reshape(-1, self.embed_dim)
        return self.norm(x)


class DotsViTPreprocessor(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()
        self.patchifier = DotsPatchEmbed(config)

    def __call__(self, x: mx.array, grid_thw=None) -> mx.array:
        return self.patchifier(x, grid_thw)


class DotsVisionBlock(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()
        self.attn = VisionAttention(config)
        self.norm1 = RMSNorm(config.embed_dim, eps=config.rms_norm_eps)
        self.mlp = DotsSwiGLUFFN(config)
        self.norm2 = RMSNorm(config.embed_dim, eps=config.rms_norm_eps)

    def __call__(
        self,
        hidden_states: mx.array,
        cu_seqlens: mx.array,
        rotary_pos_emb: mx.array,
    ) -> mx.array:
        hidden_states = hidden_states + self.attn(
            self.norm1(hidden_states),
            cu_seqlens=cu_seqlens,
            rotary_pos_emb=rotary_pos_emb,
        )
        hidden_states = hidden_states + self.mlp(self.norm2(hidden_states))
        return hidden_states


class VisionModel(nn.Module):
    def __init__(self, config: VisionConfig) -> None:
        super().__init__()
        self.config = config
        self.model_type = config.model_type
        self.spatial_merge_size = config.spatial_merge_size

        self.patch_embed = DotsViTPreprocessor(config)
        head_dim = config.embed_dim // config.num_attention_heads
        self.rotary_pos_emb = VisionRotaryEmbedding(head_dim // 2)

        self.blocks = [DotsVisionBlock(config) for _ in range(config.num_hidden_layers)]

        if self.config.post_norm:
            self.post_trunk_norm = RMSNorm(config.embed_dim, eps=config.rms_norm_eps)

        self.merger = PatchMerger(
            dim=config.hidden_size,
            context_dim=config.embed_dim,
            spatial_merge_size=config.spatial_merge_size,
        )

    def get_pos_ids_by_grid(self, grid_thw: mx.array):
        pos_ids = []
        for t, h, w in grid_thw.tolist():
            hpos_ids = mx.arange(h).reshape(h, 1)
            hpos_ids = mx.repeat(hpos_ids, w, axis=1)
            hpos_ids = hpos_ids.reshape(
                h // self.spatial_merge_size,
                self.spatial_merge_size,
                w // self.spatial_merge_size,
                self.spatial_merge_size,
            )
            hpos_ids = hpos_ids.transpose(0, 2, 1, 3).flatten()

            wpos_ids = mx.arange(w).reshape(1, w)
            wpos_ids = mx.repeat(wpos_ids, h, axis=0)
            wpos_ids = wpos_ids.reshape(
                h // self.spatial_merge_size,
                self.spatial_merge_size,
                w // self.spatial_merge_size,
                self.spatial_merge_size,
            )
            wpos_ids = wpos_ids.transpose(0, 2, 1, 3).flatten()

            pos_ids.append(mx.tile(mx.stack([hpos_ids, wpos_ids], axis=-1), (t, 1)))
        return pos_ids

    def rot_pos_emb(self, grid_thw: mx.array) -> mx.array:
        pos_ids = mx.concatenate(self.get_pos_ids_by_grid(grid_thw), axis=0)
        max_grid_size = int(mx.max(grid_thw[:, 1:]).item())
        rotary_pos_emb_full = self.rotary_pos_emb(max_grid_size)
        rotary_pos_emb = rotary_pos_emb_full[pos_ids].reshape(pos_ids.shape[0], -1)
        return rotary_pos_emb

    def __call__(
        self,
        hidden_states: mx.array,
        grid_thw: mx.array,
        output_hidden_states: Optional[bool] = None,
    ) -> mx.array:
        hidden_states = self.patch_embed(hidden_states, grid_thw)
        rotary_pos_emb = self.rot_pos_emb(grid_thw)

        cu_seqlens = []
        for i in range(grid_thw.shape[0]):
            seq_len = grid_thw[i, 1] * grid_thw[i, 2]
            cu_seqlens.append(mx.repeat(seq_len, grid_thw[i, 0]))
        cu_seqlens = mx.concatenate(cu_seqlens).astype(mx.int32)
        cu_seqlens = mx.cumsum(cu_seqlens, axis=0)
        cu_seqlens = mx.pad(cu_seqlens, (1, 0), constant_values=0)

        for block in self.blocks:
            hidden_states = block(
                hidden_states, cu_seqlens=cu_seqlens, rotary_pos_emb=rotary_pos_emb
            )

        if self.config.post_norm:
            hidden_states = self.post_trunk_norm(hidden_states)

        hidden_states = self.merger(hidden_states)
        return hidden_states

    def sanitize(self, weights):
        sanitized_weights = {}
        for key, value in weights.items():
            if "position_ids" in key:
                continue
            if "vision_tower.patch_embed.patchifier.proj.weight" in key:
                if check_array_shape(value):
                    sanitized_weights[key] = value
                else:
                    sanitized_weights[key] = value.transpose(0, 2, 3, 1)
            else:
                sanitized_weights[key] = value
        return sanitized_weights



from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from ..base import InputEmbeddingsFeatures, LanguageModelOutput
from .config import ModelConfig
from .language import LanguageModel
from .vision import VisionModel


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.language_model = LanguageModel(config.text_config, config)
        self.vision_tower = VisionModel(config.vision_config)

    @property
    def layers(self):
        return self.language_model.layers

    def build_batched_images(
        self,
        input_ids: mx.array,
        pixel_values: mx.array,
        image_token_pooling: mx.array,
        image_grids: mx.array,
        image_num_crops: mx.array,
    ) -> tuple[mx.array, mx.array]:
        raw_counts = (input_ids == self.config.image_end_token_id).sum(axis=1)
        counts = raw_counts // 2
        batch_size = counts.shape[0]

        num_images = int(counts.sum().item())

        if image_grids.shape[0] != num_images:
            raise ValueError(
                f"Expected {num_images} image grids, got {image_grids.shape[0]}"
            )
        if image_num_crops.shape[0] != num_images:
            raise ValueError(
                f"Expected {num_images} image crop counts, got {image_num_crops.shape[0]}"
            )

        num_pooled_patches_per_image = (
            (image_grids[:, :2].prod(axis=1) + image_grids[:, 2:].prod(axis=1))
            .astype(image_num_crops.dtype)
            .reshape(-1)
        )

        n_crops, n_patches, pixels_per_patch = pixel_values.shape

        example_ids_for_image = mx.array(
            np.repeat(
                np.arange(batch_size), np.array(counts).astype(np.int32).tolist()
            ),
            dtype=mx.int32,
        )

        crops_per_example = mx.zeros((batch_size,), dtype=image_num_crops.dtype)
        pooled_per_example = mx.zeros(
            (batch_size,), dtype=num_pooled_patches_per_image.dtype
        )
        for image_idx in range(num_images):
            ex = int(example_ids_for_image[image_idx].item())
            crops_per_example[ex] = crops_per_example[ex] + image_num_crops[image_idx]
            pooled_per_example[ex] = (
                pooled_per_example[ex] + num_pooled_patches_per_image[image_idx]
            )

        total_crops = int(crops_per_example.sum().item())
        if total_crops != n_crops:
            raise ValueError(f"Expected {total_crops} crops, got {n_crops}")

        total_pooled = int(pooled_per_example.sum().item())
        if total_pooled != image_token_pooling.shape[0]:
            raise ValueError(
                f"Expected {total_pooled} pooled patches, got {image_token_pooling.shape[0]}"
            )

        max_crops = int(crops_per_example.max().item())
        images = mx.full(
            (batch_size, max_crops, n_patches, pixels_per_patch),
            vals=-1,
            dtype=pixel_values.dtype,
        )

        offset_crop = 0
        for i in range(batch_size):
            num = int(crops_per_example[i].item())
            images[i, :num] = pixel_values[offset_crop : offset_crop + num]
            offset_crop += num

        max_pooled = int(pooled_per_example.max().item())
        token_dim = image_token_pooling.shape[1]
        new_token_pooling = mx.full(
            (batch_size, max_pooled, token_dim),
            vals=-1,
            dtype=image_token_pooling.dtype,
        )

        patches_per_image = image_num_crops * n_patches
        counts_list = counts.tolist()
        image_idx = 0
        pooled_offset = 0
        patch_offset = 0
        for ex, c in enumerate(counts_list):
            num_pooled = int(pooled_per_example[ex].item())
            cur = mx.array(
                image_token_pooling[pooled_offset : pooled_offset + num_pooled]
            )

            per_img_patches = patches_per_image[image_idx : image_idx + c]
            index_offsets = [0] + np.cumsum(per_img_patches.tolist()).tolist()[:-1]
            per_img_pooled = num_pooled_patches_per_image[
                image_idx : image_idx + c
            ].tolist()

            offset = 0
            for j in range(c):
                n = int(per_img_pooled[j])
                idx_off = int(index_offsets[j])
                cur_slice = cur[offset : offset + n]
                cur[offset : offset + n] = mx.where(
                    cur_slice >= 0,
                    cur_slice + idx_off,
                    cur_slice,
                )
                offset += n

            new_token_pooling[ex, :num_pooled] = cur
            pooled_offset += num_pooled
            image_idx += c
            patch_offset += num_pooled

        return images, new_token_pooling

    def build_batched_videos(
        self,
        input_ids: mx.array,
        pixel_values_videos: mx.array,
        video_token_pooling: mx.array,
        video_grids: mx.array,
    ) -> tuple[mx.array, mx.array]:
        end_token_id = (
            self.config.frame_end_token_id
            if self.config.use_frame_special_tokens
            else self.config.image_end_token_id
        )
        counts = mx.any(input_ids == end_token_id, axis=1).astype(mx.int32)
        batch_size = counts.shape[0]
        num_videos = int(counts.sum().item())

        if video_grids.shape[0] != num_videos:
            raise ValueError(
                f"Expected {num_videos} videos, got {video_grids.shape[0]}"
            )

        num_pooled_patches_per_video = (video_grids[:, 1] * video_grids[:, 2]).astype(
            video_token_pooling.dtype
        )

        n_frames, n_patches, pixels_per_patch = pixel_values_videos.shape

        frames_per_example = mx.zeros((batch_size,), dtype=mx.int32)
        pooled_per_example = mx.zeros((batch_size,), dtype=video_token_pooling.dtype)

        video_index = 0
        for i in range(batch_size):
            if counts[i].item() == 1:
                frames_per_example[i] = int(video_grids[video_index][0].item())
                pooled_per_example[i] = num_pooled_patches_per_video[video_index]
                video_index += 1

        max_frames = int(frames_per_example.max().item()) if num_videos else 0
        videos = mx.full(
            (batch_size, max_frames, n_patches, pixels_per_patch),
            vals=-1,
            dtype=pixel_values_videos.dtype,
        )

        offset = 0
        for i in range(batch_size):
            num = int(frames_per_example[i].item())
            if num > 0:
                videos[i, :num] = pixel_values_videos[offset : offset + num]
                offset += num

        max_pooled = int(pooled_per_example.max().item()) if num_videos else 0
        token_dim = video_token_pooling.shape[1]
        new_token_pooling = mx.full(
            (batch_size, max_pooled, token_dim),
            vals=-1,
            dtype=video_token_pooling.dtype,
        )

        pooled_offset = 0
        for i in range(batch_size):
            num = int(pooled_per_example[i].item())
            if num > 0:
                new_token_pooling[i, :num] = video_token_pooling[
                    pooled_offset : pooled_offset + num
                ]
                pooled_offset += num

        if offset != n_frames:
            raise ValueError(f"Expected {n_frames} frames, got {offset}")
        if pooled_offset != video_token_pooling.shape[0]:
            raise ValueError(
                f"Expected {video_token_pooling.shape[0]} pooled tokens, got {pooled_offset}"
            )

        return videos, new_token_pooling

    def merge_visual_inputs(
        self,
        *,
        input_ids: mx.array,
        pixel_values: Optional[mx.array] = None,
        image_token_pooling: Optional[mx.array] = None,
        image_grids: Optional[mx.array] = None,
        image_num_crops: Optional[mx.array] = None,
        video_token_pooling: Optional[mx.array] = None,
        video_grids: Optional[mx.array] = None,
    ) -> tuple[Optional[mx.array], Optional[mx.array]]:
        if pixel_values is None:
            return None, None

        if video_token_pooling is not None or video_grids is not None:
            if video_token_pooling is None or video_grids is None:
                raise ValueError(
                    "video_token_pooling and video_grids are required for videos"
                )
            return self.build_batched_videos(
                input_ids=input_ids,
                pixel_values_videos=pixel_values,
                video_token_pooling=video_token_pooling,
                video_grids=video_grids,
            )

        if (
            image_token_pooling is None
            or image_grids is None
            or image_num_crops is None
        ):
            raise ValueError(
                "image_token_pooling, image_grids, and image_num_crops are required for images"
            )

        return self.build_batched_images(
            input_ids=input_ids,
            pixel_values=pixel_values,
            image_token_pooling=image_token_pooling,
            image_grids=image_grids,
            image_num_crops=image_num_crops,
        )

    def get_input_embeddings(
        self,
        input_ids: mx.array,
        pixel_values: Optional[mx.array] = None,
        **kwargs,
    ) -> mx.array:
        input_ids = input_ids * (input_ids != -1).astype(input_ids.dtype)
        x = self.language_model.model.wte(input_ids)

        if pixel_values is not None:

            pixel_values, token_pooling = self.merge_visual_inputs(
                input_ids=input_ids,
                pixel_values=pixel_values,
                image_token_pooling=kwargs.get("image_token_pooling", None),
                image_grids=kwargs.get("image_grids", None),
                image_num_crops=kwargs.get("image_num_crops", None),
                video_token_pooling=kwargs.get("video_token_pooling", None),
                video_grids=kwargs.get("video_grids", None),
            )

            dtype = self.vision_tower.image_vit.patch_embedding.weight.dtype
            pixel_values = pixel_values.astype(dtype)
            image_features = self.vision_tower(pixel_values, token_pooling)
            is_image_patch = mx.reshape(input_ids, (-1,)) == self.config.image_patch_id
            if int(is_image_patch.sum().item()) != image_features.shape[0]:
                raise ValueError(
                    f"Expected {int(is_image_patch.sum().item())} image features, got {image_features.shape[0]}"
                )
            flat_x = mx.reshape(x, (-1, x.shape[-1]))
            positions = mx.array(np.where(np.array(is_image_patch))[0], dtype=mx.uint32)
            flat_x[positions] = flat_x[positions] + image_features
            x = flat_x.reshape(x.shape)

        return InputEmbeddingsFeatures(inputs_embeds=x)

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
        **kwargs,
    ) -> LanguageModelOutput:
        if input_ids.ndim == 1:
            input_ids = input_ids[None, :]

        input_embeddings_features = self.get_input_embeddings(
            input_ids=input_ids, pixel_values=pixel_values, **kwargs
        )

        return self.language_model(
            input_ids,
            inputs_embeds=input_embeddings_features.inputs_embeds,
            mask=mask,
            cache=cache,
        )

    def sanitize(self, weights):
        def transform_key(key: str) -> str:
            if key.startswith("model.transformer."):
                key = key.replace("model.transformer.", "language_model.model.", 1)
            if key.startswith("model.vision_backbone."):
                key = key.replace("model.vision_backbone.", "vision_tower.", 1)
            if key.startswith("lm_head."):
                key = key.replace("lm_head.", "language_model.lm_head.", 1)
            # Vision transformer uses list not named submodule
            key = key.replace(".transformer.resblocks.", ".transformer.")
            return key

        return {transform_key(k): v for k, v in weights.items()}



import inspect
from dataclasses import dataclass
from typing import Dict, List, Optional, Union

from ..base import BaseModelConfig


@dataclass
class TextConfig(BaseModelConfig):
    model_type: str = "dots_ocr"
    vocab_size: int = 151936
    hidden_size: int = 1536
    intermediate_size: int = 8960
    num_hidden_layers: int = 28
    num_attention_heads: int = 12
    num_key_value_heads: Optional[int] = 2
    max_position_embeddings: int = 131072
    rms_norm_eps: float = 1e-6
    rope_theta: float = 1000000.0
    rope_traditional: bool = False
    rope_scaling: Optional[Dict[str, Union[float, str]]] = None
    attention_bias: bool = True
    attention_dropout: float = 0.0
    tie_word_embeddings: bool = False
    use_cache: bool = True

    def __post_init__(self):
        if self.num_key_value_heads is None:
            self.num_key_value_heads = self.num_attention_heads

        if self.rope_scaling:
            required_keys = {"factor", "type"}
            if not all(key in self.rope_scaling for key in required_keys):
                raise ValueError(f"rope_scaling must contain keys {required_keys}")

            if self.rope_scaling["type"] != "linear":
                raise ValueError("rope_scaling 'type' currently only supports 'linear'")


@dataclass
class VisionConfig(BaseModelConfig):
    model_type: str = "dots_vit"
    embed_dim: int = 1536
    hidden_size: int = 1536
    intermediate_size: int = 4224
    num_hidden_layers: int = 42
    num_attention_heads: int = 12
    num_channels: int = 3
    patch_size: int = 14
    post_norm: bool = True
    rms_norm_eps: float = 1e-5
    spatial_merge_size: int = 2
    temporal_patch_size: int = 1
    use_bias: bool = False
    attn_implementation: str = "flash_attention_2"
    init_merger_std: float = 0.02
    initializer_range: float = 0.02
    is_causal: bool = False
    gradient_checkpointing: bool = False


@dataclass
class ModelConfig(BaseModelConfig):
    text_config: TextConfig
    vision_config: VisionConfig
    model_type: str = "dots_ocr"
    ignore_index: int = -100
    image_token_id: int = 151665
    video_token_id: int = 151656
    vocab_size: int = 151936
    eos_token_id: Optional[List[int]] = None
    quantization: Optional[Dict] = None

    @classmethod
    def from_dict(cls, params):
        if not params.get("text_config", {}):
            excluded_keys = {"vision_config"}
            params["text_config"] = dict(
                filter(lambda x: x[0] not in excluded_keys, params.items())
            )

        if not params.get("vision_config", {}).get("model_type", {}):
            params["vision_config"]["model_type"] = "dots_vit"

        return cls(
            text_config=TextConfig.from_dict(params["text_config"]),
            vision_config=VisionConfig.from_dict(params["vision_config"]),
            **{
                k: v
                for k, v in params.items()
                if k in inspect.signature(cls).parameters
                and k not in ["text_config", "vision_config"]
            },
        )
